import argparse
import csv
from dataclasses import dataclass
from datetime import datetime
from email.message import EmailMessage
from pathlib import Path
from smtplib import SMTP

import pandas as pd
import rich.markdown
import rich.progress
import rich.prompt
from jinja2 import Template
from pydantic import Field
from pydantic_settings import BaseSettings
from rich import print
from yarl import URL


class Settings(BaseSettings):
    smtp_host: str = Field(default=...)
    smtp_port: int = Field(default=...)
    smtp_user: str = Field(default=...)
    smtp_password: str = Field(default=...)
    smtp_from: str = Field(default=...)

    class Config:
        env_file = ".env"


settings = Settings()


class Constants:
    class PolyOpinion:
        firstname: str = "firstname"
        email: str = "email"
        token: str = "token"
        usesleft: str = "usesleft"

    class Filter:
        email: str = "email"
        status: str = "status"


@dataclass(frozen=True)
class Arguments:
    file: Path
    template: Path
    subject: str
    base_url: URL
    filters: list[Path]
    only: str


def get_args() -> Arguments:
    parser = argparse.ArgumentParser(
        description="Send emails to AEP members with vote link"
    )

    parser.add_argument(
        "--file",
        "-f",
        type=Path,
        help="Path to the CSV containing tokens and emails",
        required=True,
    )

    parser.add_argument(
        "--template",
        "-t",
        type=Path,
        help="Jinja2 compatible HTML template used for the email",
        required=True,
    )

    parser.add_argument(
        "--subject",
        "-s",
        type=str,
        help="Email subject",
        required=True,
    )

    parser.add_argument(
        "--base-url",
        type=URL,
        help="Base URL for the vote link, the token will be appended to this URL",
        required=True,
    )

    parser.add_argument(
        "--filter",
        action="append",
        type=Path,
        help="Path to a CSV containing emails to filter out",
        required=False,
    )

    parser.add_argument(
        "--only",
        type=str,
        help="Only send to this email",
        required=False,
    )

    args = parser.parse_args()

    return Arguments(
        file=args.file,
        template=args.template,
        subject=args.subject,
        base_url=args.base_url,
        filters=args.filter,
        only=args.only,
    )


def main() -> None:
    args = get_args()

    with rich.progress.open(
        args.file,
        "r",
        encoding="utf-8",
        description="Reading tokens...",
    ) as f:
        df = pd.read_csv(
            f,
            usecols=[
                Constants.PolyOpinion.firstname,
                Constants.PolyOpinion.email,
                Constants.PolyOpinion.token,
                Constants.PolyOpinion.usesleft,
            ],
        )

    original_len = len(df)

    print("Filtering used tokens...")
    df = df[df[Constants.PolyOpinion.usesleft] > 0]

    if args.filters:
        print("Filtering out excluded emails...")
        for filter in args.filters:
            with rich.progress.open(
                filter,
                "r",
                encoding="utf-8",
                description=f"Reading filter {filter.name}...",
            ) as f:
                filter_df = pd.read_csv(
                    f,
                    usecols=[
                        Constants.Filter.email,
                        Constants.Filter.status,
                    ],
                )

            # filter out emails with status="sent"
            filter_df_sent = filter_df[filter_df["status"] == "sent"]
            df = df[
                ~df[Constants.PolyOpinion.email].isin(
                    filter_df_sent[Constants.Filter.email]
                )
            ]

    if args.only:
        print(f"Only sending to {args.only}...")
        df = df[df[Constants.PolyOpinion.email] == args.only]

    if df.empty:
        print("[bold red]No emails to send[/bold red]")
        print("[bold red]Exiting...[bold red]")
        return

    print(rich.markdown.Markdown(df.head().to_markdown()))
    print(f"Original length: {original_len} tokens")
    print(f"Filtered length: {len(df)} tokens")

    is_ok = rich.prompt.Confirm.ask("Do you want to continue?")

    if not is_ok:
        print("[bold red]Exiting...[bold red]")
        return

    with rich.progress.open(
        args.template,
        "r",
        encoding="utf-8",
        description="Reading template...",
    ) as f:
        template = Template(f.read())

    log_file = Path(f"sent-{datetime.now().strftime('%Y%m%d-%H%M%S')}.csv")
    with (
        log_file.open("w", encoding="utf-8") as f,
        SMTP(settings.smtp_host, settings.smtp_port) as server,
    ):
        writer = csv.writer(f)
        writer.writerow([Constants.Filter.email, Constants.Filter.status])

        server.starttls()
        server.login(settings.smtp_user, settings.smtp_password)

        for _, row in rich.progress.track(
            df.iterrows(),
            description="Sending emails...",
            total=len(df),
        ):
            try:
                first_name = row[Constants.PolyOpinion.firstname]
                email = row[Constants.PolyOpinion.email]
                token = row[Constants.PolyOpinion.token]

                if not first_name or not email or not token:
                    raise ValueError(f"Missing data for {email}")

                vote_url = args.base_url.update_query(token=token)
                html_content = template.render(first_name=first_name, vote_url=vote_url)

                message = EmailMessage()
                message["From"] = settings.smtp_from
                message["To"] = email

                message["Subject"] = args.subject
                message.set_content(html_content, subtype="html")

                server.send_message(message)
                writer.writerow([email, "sent"])

            except Exception as e:
                writer.writerow([email, f"error: {e}"])


if __name__ == "__main__":
    main()
